from concurrent.futures import ThreadPoolExecutor
from multiprocessing.queues import Queue as MultiprocessingQueue
from multiprocessing.synchronize import Event as MultiprocessingEvent
from queue import Empty, ShutDown
from time import sleep, time
from typing import Callable, Optional

from .._queues import Message, Queue
from .._utils.logging import worker_logger
from .._utils.pydantic import PydanticModel
from .._utils.retry import handle_failure


class HeartBeat(PydanticModel):
    worker_id: int
    timestamp: float
    message: Optional[Message] = None


class Worker:
    worker_id: int

    queue: Queue[Message]

    shutdown: MultiprocessingEvent
    heartbeats: MultiprocessingQueue[HeartBeat]

    on_message_received: Callable[[Message], None]
    on_message_success: Callable[[Message], None]
    on_message_failure: Callable[[Message, Exception], None]

    def __init__(
        self,
        worker_id: int,
        queue: Queue[Message],
        shutdown_event: MultiprocessingEvent,
        heartbeats: MultiprocessingQueue[HeartBeat],
        on_message_received: Callable[[Message], None],
        on_message_success: Callable[[Message], None],
        on_message_failure: Callable[[Message, Exception], None],
    ) -> None:
        self.worker_id = worker_id

        self.queue = queue

        self.shutdown_event = shutdown_event
        self.heartbeats = heartbeats

        self.on_message_received = on_message_received
        self.on_message_success = on_message_success
        self.on_message_failure = on_message_failure

    def run(self) -> None:
        while not self.shutdown_event.is_set():
            message = self._get_message()

            if message:
                self._heartbeat(message)
                self.on_message_received(message)

                log_extra = {
                    "message_id": message.id,
                    "handler": message.eventide_metadata.handler.name,
                    "attempt": message.eventide_metadata.attempt,
                }

                worker_logger.info(f"Message {message.id} received", extra=log_extra)

                handler = message.eventide_metadata.handler
                with ThreadPoolExecutor(max_workers=1) as executor:
                    start = time()
                    future = executor.submit(handler, message)

                    try:
                        future.result(timeout=handler.timeout)
                    except Exception as exception:
                        end = time()

                        self._heartbeat(None)
                        handle_failure(message, self.queue, exception)
                        self.on_message_failure(message, exception)

                        if (
                            isinstance(exception, TimeoutError)
                            and end - start >= handler.timeout
                        ):
                            break
                    else:
                        end = time()

                        self._heartbeat(None)
                        self.queue.ack_message(message)
                        self.on_message_success(message)

                        worker_logger.info(
                            f"Message {message.id} handling succeeded in "
                            f"{end - start}s",
                            extra={**log_extra, "duration": end - start},
                        )
            else:
                sleep(0.1)

    def _get_message(self) -> Optional[Message]:
        try:
            return self.queue.get_message()
        except (Empty, ShutDown):
            return None

    def _heartbeat(self, message: Optional[Message] = None) -> None:
        try:
            self.heartbeats.put_nowait(
                HeartBeat(worker_id=self.worker_id, timestamp=time(), message=message),
            )
        except (Empty, ShutDown):
            pass



from logging import getLogger

eventide_logger = getLogger(name="eventide")
cron_logger = getLogger(name="eventide.cron")
queue_logger = getLogger(name="eventide.queue")
worker_logger = getLogger(name="eventide.worker")



from logging import INFO
from typing import Any, Literal

from pydantic import Field, NonNegativeInt, PositiveFloat, PositiveInt, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from .._queues import QueueConfig


class EventideConfig(BaseSettings):
    model_config = SettingsConfigDict(
        case_sensitive=False,
        extra="ignore",
        validate_default=False,
    )

    queue: QueueConfig

    min_pull_interval: PositiveFloat = 1.0
    max_pull_interval: PositiveFloat = 10.0

    concurrency: PositiveInt = 1

    timeout: PositiveFloat = 60.0

    retry_for: list[type[Exception]] = Field(default_factory=list)
    retry_min_backoff: PositiveFloat = 1.0
    retry_max_backoff: PositiveFloat = 60.0
    retry_limit: NonNegativeInt = 3

    log_level: Literal[0, 10, 20, 30, 40, 50] = INFO

    @field_validator("log_level", mode="before")
    @classmethod
    def parse_log_level(cls, value: Any) -> Any:
        try:
            return int(value)
        except (ValueError, TypeError):
            return value



from collections.abc import Iterable
from functools import wraps
from typing import Any, Callable, Optional, Union, cast

from .._handlers import Handler, HandlerMatcher, MatcherCallable
from .._queues import Message
from .config import EventideConfig


class HandlerManager:
    config: EventideConfig

    handlers: set[Handler]

    def __init__(self, config: EventideConfig) -> None:
        self.config = config
        self.handlers = set()

    def register(
        self,
        *matchers: Union[str, MatcherCallable],
        operator: Callable[[Iterable[bool]], bool] = all,
        timeout: Optional[float] = None,
        retry_for: Optional[list[type[Exception]]] = None,
        retry_limit: Optional[int] = None,
        retry_min_backoff: Optional[float] = None,
        retry_max_backoff: Optional[float] = None,
    ) -> Callable[..., Any]:
        def decorator(func: Callable[[Message], Any]) -> Handler:
            @wraps(func)
            def wrapper(message: Message) -> Any:
                return func(message)

            handler = cast(Handler, wrapper)
            handler.name = f"{func.__module__}.{func.__qualname__}"
            handler.matcher = HandlerMatcher(*matchers, operator=operator)
            handler.timeout = timeout if timeout is not None else self.config.timeout
            handler.retry_for = (
                retry_for if retry_for is not None else self.config.retry_for
            )
            handler.retry_limit = (
                retry_limit if retry_limit is not None else self.config.retry_limit
            )
            handler.retry_min_backoff = (
                retry_min_backoff
                if retry_min_backoff is not None
                else self.config.retry_min_backoff
            )
            handler.retry_max_backoff = (
                retry_max_backoff
                if retry_max_backoff is not None
                else self.config.retry_max_backoff
            )

            self.handlers.add(handler)

            return handler

        return decorator



from importlib import import_module
from multiprocessing import get_context
from multiprocessing.context import ForkContext
from multiprocessing.synchronize import Event as MultiprocessingEvent
from pathlib import Path
from signal import SIGINT, SIGTERM, signal
from sys import exit as sys_exit
from sys import path
from time import sleep, time
from types import FrameType
from typing import Any, Callable, Optional

from .._handlers import Handler
from .._queues import Message
from .._utils.logging import eventide_logger
from .config import EventideConfig
from .cron import CronManager
from .handler import HandlerManager
from .hook import HookManager
from .queue import QueueManager
from .worker import WorkerManager


class Eventide:
    config: EventideConfig

    context: ForkContext
    shutdown_event: MultiprocessingEvent

    handler_manager: HandlerManager
    hook_manager: HookManager
    queue_manager: QueueManager
    worker_manager: WorkerManager

    def __init__(self, config: EventideConfig) -> None:
        self.config = config

        self.context = get_context("fork")
        self.shutdown_event = self.context.Event()

        self.handler_manager = HandlerManager(config=self.config)
        self.hook_manager = HookManager()
        self.queue_manager = QueueManager(
            config=self.config,
            context=self.context,
            handler_manager=self.handler_manager,
        )
        self.worker_manager = WorkerManager(
            config=self.config,
            context=self.context,
            shutdown_event=self.shutdown_event,
            hook_manager=self.hook_manager,
            queue_manager=self.queue_manager,
        )
        self.cron_manager = CronManager(queue_manager=self.queue_manager)

        self.autodiscover()

    @property
    def handler(self) -> Callable[..., Callable[..., Handler]]:
        return self.handler_manager.register

    @property
    def cron(self) -> Callable[[str], Callable[[Callable[[], Any]], Callable[[], Any]]]:
        return self.cron_manager.register

    def on_start(self, hook: Callable[[], None]) -> Callable[[], None]:
        self.hook_manager.register_start_hook(hook)
        return hook

    def on_shutdown(self, hook: Callable[[], None]) -> Callable[[], None]:
        self.hook_manager.register_shutdown_hook(hook)
        return hook

    def on_message_received(
        self,
        hook: Callable[[Message], None],
    ) -> Callable[[Message], None]:
        self.hook_manager.register_message_received_hook(hook)
        return hook

    def on_message_success(
        self,
        hook: Callable[[Message], None],
    ) -> Callable[[Message], None]:
        self.hook_manager.register_message_success_hook(hook)
        return hook

    def on_message_failure(
        self,
        hook: Callable[[Message, Exception], None],
    ) -> Callable[[Message, Exception], None]:
        self.hook_manager.register_message_failure_hook(hook)
        return hook

    def run(self) -> None:
        eventide_logger.info("Starting Eventide...")

        self.shutdown_event.clear()
        self.setup_signal_handlers()
        self.autodiscover()
        self.hook_manager.on_start()
        self.queue_manager.start()
        self.worker_manager.start()

        while not self.shutdown_event.is_set():
            self.queue_manager.enqueue_retries()
            self.queue_manager.enqueue_messages()

            interval_start = time()
            while (
                time() - interval_start < self.queue_manager.pull_interval
                and not self.worker_manager.shutdown_event.is_set()
            ):
                self.worker_manager.monitor_workers()
                sleep(0.1)

        eventide_logger.info("Stopping Eventide...")

        self.shutdown(force=False)

    def run_cron(self) -> None:
        eventide_logger.info("Starting Eventide cron...")

        self.shutdown_event.clear()
        self.setup_signal_handlers()
        self.autodiscover()
        self.queue_manager.start()

        while not self.shutdown_event.is_set():
            self.cron_manager.evaluate_crons()
            sleep(1.0)

        self.shutdown(force=False)

    def setup_signal_handlers(self) -> None:
        def handle_signal(_signum: int, _frame: Optional[FrameType]) -> None:
            if not self.worker_manager.shutdown_event.is_set():
                eventide_logger.info("Shutting down gracefully...")
                self.shutdown_event.set()
            else:
                eventide_logger.info("Forcing immediate shutdown...")
                self.shutdown(force=True)
                sys_exit(1)

        signal(SIGINT, handle_signal)
        signal(SIGTERM, handle_signal)

    def autodiscover(self) -> None:
        cwd = Path(".").resolve()
        ignored_dirs = {".git", ".venv", "venv", "__pycache__"}

        if str(cwd) not in path:
            path.insert(0, str(cwd))

        for python_file in cwd.glob("*.py"):
            try:
                import_module(python_file.stem)
            except (ImportError, ModuleNotFoundError, TypeError):
                pass

        def import_from_directory(directory: Path) -> None:
            if not (directory / "__init__.py").exists():
                return

            for item in directory.iterdir():
                if item.is_dir() and item.name not in ignored_dirs:
                    import_from_directory(item)
                elif item.suffix == ".py":
                    module_name = ".".join(item.relative_to(cwd).with_suffix("").parts)

                    try:
                        import_module(module_name)
                    except (ImportError, ModuleNotFoundError, TypeError):
                        pass

        for child in cwd.iterdir():
            if child.is_dir() and child.name not in ignored_dirs:
                import_from_directory(child)

    def shutdown(self, force: bool = False) -> None:
        self.shutdown_event.set()

        self.worker_manager.shutdown(force=force)
        self.queue_manager.shutdown()

        self.hook_manager.on_shutdown()



from multiprocessing.context import ForkContext
from queue import Empty
from time import time

from .._queues import Message, Queue
from .._utils.logging import eventide_logger
from .config import EventideConfig
from .handler import HandlerManager


class QueueManager:
    config: EventideConfig
    context: ForkContext

    handler_manager: HandlerManager

    queue: Queue[Message]

    _empty_pulls: int

    def __init__(
        self,
        config: EventideConfig,
        context: ForkContext,
        handler_manager: HandlerManager,
    ) -> None:
        self.config = config
        self.context = context

        self.handler_manager = handler_manager

    @property
    def pull_interval(self) -> float:
        return float(
            min(
                self.config.max_pull_interval,
                self.config.min_pull_interval * (2**self._empty_pulls),
            )
        )

    def start(self) -> None:
        self.queue = Queue.factory(config=self.config.queue, context=self.context)
        self._empty_pulls = 0

    def shutdown(self) -> None:
        if hasattr(self, "queue"):
            self.queue.shutdown()

    def enqueue_retries(self) -> None:
        messages = []

        while True:
            try:
                messages.append(self.queue.get_retry_message())
            except Empty:
                break

        for message in sorted(messages, key=lambda m: m.eventide_metadata.retry_at):
            if message.eventide_metadata.retry_at <= time() and not self.queue.full:
                self.queue.put_message(message)
            else:
                self.queue.put_retry_message(message)

    def enqueue_messages(self) -> None:
        if self.queue.should_pull:
            for message in self.queue.pull_messages():
                for handler in self.handler_manager.handlers:
                    if handler.matcher(message):
                        message.eventide_metadata.handler = handler
                        self.queue.put_message(message)
                        break

                if not message.eventide_metadata.handler:
                    eventide_logger.error(
                        f"No handler found for message {message.id}",
                        extra={"message_id": message.id},
                    )

        if self.queue.empty:
            self._empty_pulls += 1
        else:
            self._empty_pulls = 0



from datetime import datetime, timedelta, timezone
from traceback import format_tb
from typing import Any, Callable

from croniter import croniter

from .._utils.logging import cron_logger
from .._utils.pydantic import BaseModel
from .queue import QueueManager


class Cron(BaseModel):
    expression: str
    get_message_body: Callable[[], Any]


class CronManager:
    crons: dict[str, Cron]
    queue_manager: QueueManager
    _last_evaluation: datetime

    def __init__(self, queue_manager: QueueManager) -> None:
        self.crons = {}
        self.queue_manager = queue_manager
        self._last_evaluation = datetime.now(tz=timezone.utc).replace(
            microsecond=0
        ) - timedelta(seconds=1)

    def register(
        self,
        expression: str,
    ) -> Callable[[Callable[[], Any]], Callable[[], Any]]:
        def wrapper(get_message_body: Callable[[], Any]) -> Callable[[], Any]:
            self.crons[
                f"{get_message_body.__module__}.{get_message_body.__qualname__}"
            ] = Cron(
                expression=expression,
                get_message_body=get_message_body,
            )
            return get_message_body

        return wrapper

    def evaluate_crons(self) -> None:
        now = datetime.now(tz=timezone.utc).replace(microsecond=0)

        second = self._last_evaluation + timedelta(seconds=1)
        while second <= now:
            for cron in self.crons.values():
                cron_schedule = croniter(cron.expression, second - timedelta(seconds=1))

                if cron_schedule.get_next(datetime) == second:
                    get_body = cron.get_message_body

                    try:
                        self.queue_manager.queue.send_message(body=get_body())
                    except Exception as exception:
                        cron_logger.warning(
                            f"Failed to send cron message for cron "
                            f"{get_body.__module__}.{get_body.__qualname__}: "
                            f"{exception}",
                            extra={
                                "cron": cron.get_message_body,
                                "expression": cron.expression,
                                "exception": str(exception),
                                "traceback": "".join(
                                    format_tb(exception.__traceback__),
                                ),
                            },
                        )

            second += timedelta(seconds=1)

        self._last_evaluation = now



from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

from .matcher import HandlerMatcher

if TYPE_CHECKING:
    from .._queues import Message


@runtime_checkable
class Handler(Protocol):
    name: str
    matcher: HandlerMatcher
    timeout: float
    retry_for: list[type[Exception]]
    retry_limit: int
    retry_min_backoff: float
    retry_max_backoff: float

    def __call__(self, message: "Message") -> Any: ...



from typing import Any

from pydantic import Field, PositiveInt

from .queue import Message, Queue, QueueConfig


class SQSMessage(Message):
    receipt_handle: str
    attributes: dict[str, Any]
    message_attributes: dict[str, Any]


class SQSQueueConfig(QueueConfig):
    region: str
    url: str
    max_number_of_messages: PositiveInt = Field(10, le=10)
    visibility_timeout: PositiveInt = Field(30, le=12 * 60 * 60)


@Queue.register(SQSQueueConfig)
class SQSQueue(Queue[SQSMessage]):
    config: SQSQueueConfig

    def initialize(self) -> None:
        try:
            from boto3 import client
        except ImportError:
            raise ImportError(
                "Missing SQS dependencies... Install with: pip install eventide[sqs]"
            ) from None

        self.sqs_client = client("sqs", region_name=self.config.region)

    @property
    def max_messages_per_pull(self) -> int:
        return self.config.max_number_of_messages

    def send_message(self, body: Any) -> None:
        self.sqs_client.send_message(
            QueueUrl=self.config.url,
            MessageBody=self.dump_message_body(body),
        )

    def pull_messages(self) -> list[SQSMessage]:
        response = self.sqs_client.receive_message(
            QueueUrl=self.config.url,
            MaxNumberOfMessages=self.max_messages_per_pull,
            WaitTimeSeconds=1,
            VisibilityTimeout=self.config.visibility_timeout,
            AttributeNames=["All"],
            MessageAttributeNames=["All"],
            MessageSystemAttributeNames=["All"],
        )

        return [
            SQSMessage(
                id=message["MessageId"],
                body=self.load_message_body(message["Body"]),
                receipt_handle=message["ReceiptHandle"],
                attributes=message["Attributes"],
                message_attributes=message.get("MessageAttributes") or {},
            )
            for message in response.get("Messages") or []
        ]

    def ack_message(self, message: SQSMessage) -> None:
        self.sqs_client.delete_message(
            QueueUrl=self.config.url,
            ReceiptHandle=message.receipt_handle,
        )



from importlib import import_module
from os import getcwd
from sys import modules, path
from typing import TYPE_CHECKING, cast

from typer import Exit, echo

if TYPE_CHECKING:
    from .._eventide import Eventide


def resolve_app(app: str, reload: bool = False) -> "Eventide":
    if getcwd() not in path:
        path.insert(0, getcwd())

    module_name, *attrs = app.split(":", 1)
    if reload:
        root = module_name.split(".")[0]

        for sys_module in list(modules):
            if sys_module == root or sys_module.startswith(root + "."):
                del modules[sys_module]

    try:
        module = import_module(module_name)
    except ImportError:
        raise ImportError(f"Module '{module_name}' not found") from None

    for attr in [*attrs, "app", "application"]:
        if hasattr(module, attr):
            return cast("Eventide", getattr(module, attr))

    raise ValueError(f"No Eventide instance found for '{app}'")


def run_with_reload(app: str, command: str) -> None:
    try:
        from watchdog.events import FileSystemEvent, FileSystemEventHandler
        from watchdog.observers import Observer
    except ImportError:
        echo("Missing watch dependencies... Install with: pip install eventide[watch]")
        raise Exit(1) from None

    eventide_app, should_reload = resolve_app(app, reload=True), False

    class Handler(FileSystemEventHandler):
        def on_any_event(self, event: FileSystemEvent) -> None:
            nonlocal should_reload

            if str(event.src_path).endswith(".py"):
                eventide_app.shutdown(force=True)
                should_reload = True
                echo("\nChanges detected, reloading...\n")

    observer = Observer()
    observer.schedule(Handler(), ".", recursive=True)
    observer.start()

    while True:
        if command == "run":
            eventide_app.run()
        elif command == "cron":
            eventide_app.run_cron()
        else:
            raise ValueError(f"Unknown command: {command}")

        if not should_reload:
            break

        eventide_app, should_reload = resolve_app(app, reload=True), False



from typing import Callable

from .._queues import Message


class HookManager:
    _on_start_hooks: dict[str, Callable[[], None]]
    _on_shutdown_hooks: dict[str, Callable[[], None]]
    _on_message_received_hooks: dict[str, Callable[[Message], None]]
    _on_message_success_hooks: dict[str, Callable[[Message], None]]
    _on_message_failure_hooks: dict[str, Callable[[Message, Exception], None]]

    def __init__(self) -> None:
        self._on_start_hooks = {}
        self._on_shutdown_hooks = {}
        self._on_message_received_hooks = {}
        self._on_message_success_hooks = {}
        self._on_message_failure_hooks = {}

    def register_start_hook(self, hook: Callable[[], None]) -> None:
        self._on_start_hooks[f"{hook.__module__}.{hook.__qualname__}"] = hook

    def register_shutdown_hook(self, hook: Callable[[], None]) -> None:
        self._on_shutdown_hooks[f"{hook.__module__}.{hook.__qualname__}"] = hook

    def register_message_received_hook(self, hook: Callable[[Message], None]) -> None:
        self._on_message_received_hooks[f"{hook.__module__}.{hook.__qualname__}"] = hook

    def register_message_success_hook(self, hook: Callable[[Message], None]) -> None:
        self._on_message_success_hooks[f"{hook.__module__}.{hook.__qualname__}"] = hook

    def register_message_failure_hook(
        self,
        hook: Callable[[Message, Exception], None],
    ) -> None:
        self._on_message_failure_hooks[f"{hook.__module__}.{hook.__qualname__}"] = hook

    def on_start(self) -> None:
        for hook in self._on_start_hooks.values():
            hook()

    def on_shutdown(self) -> None:
        for hook in self._on_shutdown_hooks.values():
            hook()

    def on_message_received(self, message: Message) -> None:
        for hook in self._on_message_received_hooks.values():
            hook(message)

    def on_message_success(self, message: Message) -> None:
        for hook in self._on_message_success_hooks.values():
            hook(message)

    def on_message_failure(self, message: Message, exception: Exception) -> None:
        for hook in self._on_message_failure_hooks.values():
            hook(message, exception)



from collections.abc import Generator
from contextlib import contextmanager
from logging import getLogger
from typing import Any, Optional

from pydantic import Field, PositiveInt

from .queue import Message, Queue, QueueConfig


class CloudflareMessage(Message):
    lease_id: str
    metadata: dict[str, Any]
    timestamp_ms: Optional[float] = None
    attempts: Optional[float] = None


class CloudflareQueueConfig(QueueConfig):
    account_id: str
    queue_id: str
    batch_size: PositiveInt = Field(10, le=100)
    visibility_timeout_ms: PositiveInt = Field(30000, le=12 * 60 * 60 * 1000)


@Queue.register(CloudflareQueueConfig)
class CloudflareQueue(Queue[CloudflareMessage]):
    config: CloudflareQueueConfig

    @property
    def max_messages_per_pull(self) -> int:
        return self.config.batch_size

    def initialize(self) -> None:
        try:
            from cloudflare import Cloudflare
        except ImportError:
            raise ImportError(
                "Missing cloudflare dependencies... Install with: pip install "
                "eventide[cloudflare]"
            ) from None

        self.cloudflare_client = Cloudflare()

    def send_message(self, body: Any) -> None:
        # TODO: Seems like cloudflare does not support sending messages at the moment
        # with their SDK
        pass

    def pull_messages(self) -> list[CloudflareMessage]:
        with self._suppress_httpx_info_logs():
            response = self.cloudflare_client.queues.messages.pull(
                self.config.queue_id,
                account_id=self.config.account_id,
                batch_size=self.max_messages_per_pull,
            )

        return [
            CloudflareMessage(
                id=message["id"],
                body=self.load_message_body(message["body"]),
                lease_id=message["lease_id"],
                metadata=message["metadata"],
                timestamp_ms=message["timestamp_ms"],
                attempts=int(message["attempts"]),
            )
            for message in response.result["messages"]  # type: ignore[call-overload]
        ]

    def ack_message(self, message: CloudflareMessage) -> None:
        with self._suppress_httpx_info_logs():
            self.cloudflare_client.queues.messages.ack(
                self.config.queue_id,
                account_id=self.config.account_id,
                acks=[{"lease_id": message.lease_id}],
            )

    @contextmanager
    def _suppress_httpx_info_logs(self) -> Generator[None, None, None]:
        httpx_logger = getLogger("httpx")
        httpcore_logger = getLogger("httpcore")
        httpx_level = httpx_logger.level
        httpcore_level = httpcore_logger.level

        httpx_logger.setLevel("WARNING")
        httpcore_logger.setLevel("WARNING")

        try:
            yield
        finally:
            httpx_logger.setLevel(httpx_level)
            httpcore_logger.setLevel(httpcore_level)



from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Callable, Union, cast

from jmespath import compile as jmespath_compile
from jmespath.exceptions import JMESPathError

if TYPE_CHECKING:
    from .._queues import Message


MatcherCallable = Callable[[dict[str, Any]], bool]


class HandlerMatcher:
    matchers: list[MatcherCallable]
    operator: Callable[[Iterable[bool]], bool]

    def __init__(
        self,
        *matchers: Union[str, MatcherCallable],
        operator: Callable[[Iterable[bool]], bool] = all,
    ):
        if not matchers:
            raise ValueError("At least one matcher must be provided")

        self.matchers = [self._prepare_matchers(matcher) for matcher in matchers]
        self.operator = operator

    def __call__(self, message: "Message") -> bool:
        message_dict = message.model_dump()

        return self.operator(
            self._match(matcher, message_dict) for matcher in self.matchers
        )

    def _prepare_matchers(
        self,
        matcher: Union[str, MatcherCallable],
    ) -> MatcherCallable:
        if callable(matcher):
            return matcher

        try:
            return cast(MatcherCallable, jmespath_compile(matcher).search)
        except JMESPathError:
            raise ValueError(f"Invalid JMESPath expression: {matcher}") from None

    def _match(self, matcher: MatcherCallable, message_dict: dict[str, Any]) -> bool:
        try:
            return bool(matcher(message_dict))
        except JMESPathError:
            return False



from logging import INFO, basicConfig
from os import environ
from random import uniform
from time import sleep

from eventide import CloudflareQueueConfig, Eventide, EventideConfig, Message

basicConfig(level=INFO)

app = Eventide(
    config=EventideConfig(
        queue=CloudflareQueueConfig(
            account_id=environ.get("CLOUDFLARE_ACCOUNT_ID"),
            queue_id=environ.get("CLOUDFLARE_QUEUE_ID"),
            buffer_size=20,
        ),
        concurrency=2,
        timeout=2.0,
        retry_for=[Exception],
        retry_limit=2,
    ),
)


@app.handler("length(body.value) >= `1` && length(body.value) <= `5`")
def handle_1_to_5(message: Message) -> None:
    sleep(uniform(0, len(message.body["value"]) / 3.0))


@app.handler("length(body.value) >= `6` && length(body.value) <= `10`")
def handle_6_to_10(message: Message) -> None:
    sleep(uniform(0, len(message.body["value"]) / 3.0))


@app.handler(lambda message: isinstance(message["body"], str))
def handle_non_json(message: Message) -> None:
    sleep(uniform(0, len(message.body) / 3.0))



from ._eventide import Eventide, EventideConfig
from ._exceptions import EventideError, WorkerCrashedError, WorkerError
from ._queues import (
    CloudflareQueueConfig,
    Message,
    MockQueueConfig,
    Queue,
    QueueConfig,
    SQSQueueConfig,
)

__all__ = [
    "CloudflareQueueConfig",
    "Eventide",
    "EventideConfig",
    "EventideError",
    "Message",
    "MockQueueConfig",
    "Queue",
    "QueueConfig",
    "SQSQueueConfig",
    "WorkerCrashedError",
    "WorkerError",
]


for name in __all__:
    locals()[name].__module__ = "eventide"



from time import time
from traceback import format_tb
from typing import TYPE_CHECKING

from .logging import worker_logger

if TYPE_CHECKING:
    from .._handlers import Handler
    from .._queues import Message, Queue


def should_retry(handler: "Handler", attempt: int, exception: BaseException) -> bool:
    return attempt <= handler.retry_limit and any(
        isinstance(exception, exception_type) for exception_type in handler.retry_for
    )


def handle_failure(
    message: "Message",
    queue: "Queue[Message]",
    exception: Exception,
) -> None:
    handler = message.eventide_metadata.handler
    attempt = message.eventide_metadata.attempt

    exception_type = type(exception).__name__
    if should_retry(handler=handler, attempt=attempt, exception=exception):
        backoff = min(
            handler.retry_max_backoff,
            handler.retry_min_backoff * 2 ** (attempt - 1),
        )

        message.eventide_metadata.attempt = attempt + 1
        message.eventide_metadata.retry_at = time() + backoff

        queue.put_retry_message(message=message)

        worker_logger.warning(
            f"Message {message.id} handling failed with {exception_type}. Retrying in "
            f"{backoff}s",
            extra={
                "message_id": message.id,
                "handler": handler.name,
                "attempt": attempt,
                "exception": str(exception),
                "traceback": "".join(format_tb(exception.__traceback__)),
            },
        )
    else:
        worker_logger.error(
            f"Message {message.id} handling failed with {exception_type}",
            extra={
                "message_id": message.id,
                "handler": handler.name,
                "attempt": attempt,
                "exception": str(exception),
                "traceback": "".join(format_tb(exception.__traceback__)),
            },
        )



from multiprocessing.context import ForkContext, ForkProcess
from multiprocessing.queues import Queue as MultiprocessingQueue
from multiprocessing.synchronize import Event as MultiprocessingEvent
from queue import Empty
from signal import SIG_IGN, SIGINT, SIGTERM, signal
from time import time
from typing import Optional

from .._exceptions import WorkerCrashedError
from .._queues import Message
from .._utils.pydantic import PydanticModel
from .._utils.retry import handle_failure
from .._workers import HeartBeat, Worker
from .config import EventideConfig
from .hook import HookManager
from .queue import QueueManager


class WorkerState(PydanticModel):
    worker_id: int
    process: ForkProcess
    heartbeat: float
    message: Optional[Message] = None


class WorkerManager:
    config: EventideConfig
    context: ForkContext
    shutdown_event: MultiprocessingEvent

    queue_manager: QueueManager

    _heartbeats: MultiprocessingQueue[HeartBeat]
    _workers: dict[int, WorkerState]

    def __init__(
        self,
        config: EventideConfig,
        context: ForkContext,
        shutdown_event: MultiprocessingEvent,
        hook_manager: HookManager,
        queue_manager: QueueManager,
    ) -> None:
        self.config = config
        self.context = context
        self.shutdown_event = shutdown_event

        self.hook_manager = hook_manager
        self.queue_manager = queue_manager

    def start(self) -> None:
        self._heartbeats = self.context.Queue()

        self._workers = {}
        for worker_id in range(1, self.config.concurrency + 1):
            self.spawn_worker(worker_id=worker_id)

    def shutdown(self, force: bool = False) -> None:
        if hasattr(self, "_workers"):
            if not force:
                while self._workers:
                    self.monitor_workers()

            for worker_id in list(self._workers.keys()):
                self.kill_worker(worker_id=worker_id)

        if hasattr(self, "_heartbeats"):
            self._heartbeats.close()
            self._heartbeats.cancel_join_thread()

    def monitor_workers(self) -> None:
        while True:
            try:
                heartbeat = self._heartbeats.get_nowait()
            except Empty:
                break

            self._workers[heartbeat.worker_id] = WorkerState(
                worker_id=heartbeat.worker_id,
                process=self._workers[heartbeat.worker_id].process,
                heartbeat=heartbeat.timestamp,
                message=heartbeat.message,
            )

        for worker_id, worker_state in list(self._workers.items()):
            if not worker_state.process.is_alive():
                self.kill_worker(worker_id=worker_id)

                if not self.shutdown_event.is_set():
                    self.spawn_worker(worker_id=worker_id)

                if worker_state.message:
                    handle_failure(
                        worker_state.message,
                        self.queue_manager.queue,
                        WorkerCrashedError(
                            f"Worker {worker_id} crashed while handling message "
                            f"{worker_state.message.id}",
                        ),
                    )

    def spawn_worker(self, worker_id: int) -> None:
        def worker_process() -> None:
            signal(SIGINT, SIG_IGN)
            signal(SIGTERM, SIG_IGN)

            Worker(
                worker_id=worker_id,
                queue=self.queue_manager.queue,
                shutdown_event=self.shutdown_event,
                heartbeats=self._heartbeats,
                on_message_received=self.hook_manager.on_message_received,
                on_message_success=self.hook_manager.on_message_success,
                on_message_failure=self.hook_manager.on_message_failure,
            ).run()

        self._workers[worker_id] = WorkerState(
            worker_id=worker_id,
            process=self.context.Process(target=worker_process, daemon=True),
            heartbeat=time(),
            message=None,
        )
        self._workers[worker_id].process.start()

    def kill_worker(self, worker_id: int) -> None:
        worker = self._workers.pop(worker_id, None)

        if worker:
            if worker.process.is_alive():
                worker.process.terminate()

            if worker.process.is_alive():
                worker.process.kill()

            worker.process.join()



from typer import Option

from .eventide import cli
from .utils import resolve_app, run_with_reload


@cli.command("run")
def run(
    app: str = Option(..., "--app", "-a", help="App in module:attribute format"),
    reload: bool = Option(False, "--reload", "-r", help="Reload on code changes"),
) -> None:
    resolve_app(app).run() if not reload else run_with_reload(app, "run")



from random import choices, randint
from string import ascii_letters, digits
from typing import Any
from uuid import uuid4

from pydantic import NonNegativeInt, PositiveInt

from .._utils.logging import queue_logger
from .queue import Message, Queue, QueueConfig


class MockMessage(Message):
    pass


class MockQueueConfig(QueueConfig):
    min_messages: NonNegativeInt = 0
    max_messages: PositiveInt = 10


@Queue.register(MockQueueConfig)
class MockQueue(Queue[MockMessage]):
    config: MockQueueConfig

    @property
    def max_messages_per_pull(self) -> int:
        return self.config.max_messages

    def initialize(self) -> None:
        pass

    def send_message(self, body: Any) -> None:
        pass

    def pull_messages(self) -> list[MockMessage]:
        message_count = randint(self.config.min_messages, self.config.max_messages)

        queue_logger.debug(f"Pulled {message_count} messages from Mock Queue")

        return [
            self._build_json_message() if index % 2 == 0 else self._build_str_message()
            for index in range(message_count)
        ]

    def ack_message(self, message: MockMessage) -> None:
        queue_logger.debug(f"Acknowledged message {message.id}")

    def _build_json_message(self) -> MockMessage:
        return MockMessage(
            id=str(uuid4()),
            body={
                "value": "".join(choices(ascii_letters + digits, k=randint(1, 10))),
            },
        )

    def _build_str_message(self) -> MockMessage:
        return MockMessage(
            id=str(uuid4()),
            body="".join(choices(ascii_letters + digits, k=randint(1, 10))),
        )



from abc import ABC, abstractmethod
from multiprocessing.context import ForkContext
from multiprocessing.queues import Queue as MultiprocessingQueue
from multiprocessing.sharedctypes import Synchronized
from typing import Any, Callable, ClassVar, Generic, TypeVar

from orjson import JSONDecodeError, JSONEncodeError, dumps, loads
from pydantic import Field, NonNegativeInt, PositiveInt

from .._handlers import Handler
from .._utils.pydantic import PydanticModel

TMessage = TypeVar("TMessage", bound="Message")


class MessageMetadata(PydanticModel):
    attempt: PositiveInt = 1
    retry_at: float = Field(None, validate_default=False)  # type: ignore[assignment]
    handler: Handler = Field(None, validate_default=False)  # type: ignore[assignment]


class Message(PydanticModel):
    id: str
    body: Any
    eventide_metadata: MessageMetadata = Field(default_factory=MessageMetadata)


class QueueConfig(PydanticModel):
    buffer_size: NonNegativeInt = 0


class Queue(Generic[TMessage], ABC):
    queue_type_registry: ClassVar[dict[type[QueueConfig], type["Queue[Any]"]]] = {}

    config: QueueConfig
    context: ForkContext

    message_buffer: MultiprocessingQueue[TMessage]
    retry_buffer: MultiprocessingQueue[TMessage]

    _size: Synchronized  # type: ignore[type-arg]

    def __init__(self, config: QueueConfig, context: ForkContext) -> None:
        self.config = config
        self.context = context

        self.message_buffer = self.context.Queue(maxsize=self.config.buffer_size)
        self.retry_buffer = self.context.Queue()

        self._size = self.context.Value("i", 0)

        self.initialize()

    @property
    @abstractmethod
    def max_messages_per_pull(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def initialize(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def send_message(self, body: Any) -> None:
        raise NotImplementedError

    @abstractmethod
    def pull_messages(self) -> list[TMessage]:
        raise NotImplementedError

    @abstractmethod
    def ack_message(self, message: TMessage) -> None:
        raise NotImplementedError

    @classmethod
    def register(
        cls,
        queue_config_type: type[QueueConfig],
    ) -> Callable[[type["Queue[Any]"]], type["Queue[Any]"]]:
        def inner(queue_subclass: type[Queue[Any]]) -> type[Queue[Any]]:
            cls.queue_type_registry[queue_config_type] = queue_subclass
            return queue_subclass

        return inner

    @classmethod
    def factory(cls, config: QueueConfig, context: ForkContext) -> "Queue[Any]":
        queue_subclass = cls.queue_type_registry.get(type(config))

        if not queue_subclass:
            raise ValueError(
                f"No queue implementation found for {type(config).__name__}",
            )

        return queue_subclass(config=config, context=context)

    @staticmethod
    def load_message_body(body: str) -> Any:
        try:
            return loads(body)
        except JSONDecodeError:
            return body

    @staticmethod
    def dump_message_body(body: Any) -> str:
        try:
            return dumps(body).decode("utf-8")
        except JSONEncodeError:
            return str(body)

    @property
    def empty(self) -> bool:
        with self._size.get_lock():
            return bool(self._size.value == 0)

    @property
    def full(self) -> bool:
        buffer_size = self.config.buffer_size

        if buffer_size == 0:
            return False

        with self._size.get_lock():
            return bool(self._size.value == buffer_size)

    @property
    def should_pull(self) -> bool:
        buffer_size = self.config.buffer_size

        if buffer_size == 0:
            return True

        with self._size.get_lock():
            return bool(buffer_size - self._size.value >= self.max_messages_per_pull)

    def get_message(self) -> TMessage:
        message = self.message_buffer.get_nowait()

        with self._size.get_lock():
            self._size.value -= 1

        return message

    def put_message(self, message: TMessage) -> None:
        with self._size.get_lock():
            self.message_buffer.put_nowait(message)
            self._size.value += 1

    def get_retry_message(self) -> TMessage:
        return self.retry_buffer.get_nowait()

    def put_retry_message(self, message: TMessage) -> None:
        self.retry_buffer.put_nowait(message)

    def shutdown(self) -> None:
        self.message_buffer.close()
        self.message_buffer.cancel_join_thread()

        self.retry_buffer.close()
        self.retry_buffer.cancel_join_thread()



from typer import Option, echo

from .eventide import cli
from .utils import resolve_app


@cli.command("handlers")
def handlers(
    app: str = Option(
        ...,
        "--app",
        "-a",
        help="App in module:attr format, e.g. main:app",
    ),
) -> None:
    for handler in resolve_app(app).handler_manager.handlers:
        echo(handler.name)
